/*	--*- c -*--
 * Copyright (C) 2009 Enrico Scholz <enrico.scholz@informatik.tu-chemnitz.de>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; version 3 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include "dhcpd.h"
#include <string.h>
#include <stdio.h>

enum dhcpd_option_type {
	TYPE_IP,
	TYPE_STRING,
	TYPE_NET,
	TYPE_INT32,
	TYPE_INT16,
	TYPE_INT8,
	TYPE_RAW,

	TYPE_SPECIAL,
	TYPE_MAX
};

struct dhcpd_key_info;

typedef bool		(*func_filter_t)(struct dhcpd_param const *param);
typedef ssize_t		(*func_fill_t)(struct dhcpd_param const *param,
				       void *buf, size_t buf_len);
typedef bool		(*func_fill_info_t)(struct dhcpd_param *param,
					    struct dhcpd_key_info const *info,
					    char const *val);

struct dhcpd_func_set {
	func_filter_t		filter;
	func_fill_t		fill;
	func_fill_info_t	fill_info;
};

struct dhcpd_key_info {
	struct dhcpd_func_set	funcs;

	size_t			info_len;
	unsigned char		code;
	char			*vendor;
};

struct dhcpd_key_str {
	char const			*vendor;
	char const			*name;
	unsigned char			code;
	enum dhcpd_option_type		type;
	bool				is_list;
	struct dhcpd_func_set const	*funcs;
};

struct dhcpd_type_str {
	char const		*name;
	enum dhcpd_option_type	type;
};

static struct dhcpd_func_set const	FUNCS_NORM[] = {
};

static struct dhcpd_func_set const	FUNCS_LIST[] = {
};

static struct dhcpd_key_str const	DHCPD_KEYS[] = {
	{ NULL, "subnet-mask",   1, TYPE_IP,    false, NULL },
	{ NULL, "time-offset",   2, TYPE_INT32, false, NULL },
	{ NULL, "routers",       3, TYPE_IP,    true,  NULL },
};

static struct dhcpd_type_str const	DHCPD_TYPES[] = {
	{ "ip",     TYPE_IP },
	{ "string", TYPE_STRING },
	{ "net",    TYPE_NET },
	{ "int32",  TYPE_INT32 },
	{ "int16",  TYPE_INT16 },
	{ "int8",   TYPE_INT8 },
	{ "raw",    TYPE_RAW },
};

static struct dhcpd_key_str const * dhcpd_parse_key_str(char const *vendor, char *key)
{
	size_t			i;

	for (i = 0; i < ARRAY_SIZE(DHCPD_KEYS); ++i) {
		struct dhcpd_key_str const	*k = DHCPD_KEYS + i;

		if ((k->vendor == NULL || vendor == NULL) &&
		    (k->vendor != vendor))
			continue;

		if (vendor && strcmp(k->vendor, vendor) != 0)
			continue;

		if (strcmp(k->name, key) != 0)
			continue;

		return k;
	}

	fprintf(stderr, "could not parse DHCP key '%s'\n", key);
	return NULL;
}

static struct dhcpd_key_str const * dhcpd_parse_key_int(char const *vendor, unsigned long key)
{
	size_t			i;

	for (i = 0; i < ARRAY_SIZE(DHCPD_KEYS); ++i) {
		struct dhcpd_key_str const	*k = DHCPD_KEYS + i;

		if ((k->vendor == NULL || vendor == NULL) &&
		    (k->vendor != vendor))
			continue;

		if (vendor && strcmp(k->vendor, vendor) != 0)
			continue;

		if (k->code != key)
			continue;

		return k;
	}

	fprintf(stderr, "unknown numeric key '%lu'\n", key);
	return NULL;
}

static bool dhcpd_parse_type(struct dhcpd_key_str *res, char *type)
{
	size_t		len = strlen(type);
	size_t		i;

	if (len <= 2 || type[len-2] != '[' || type[len-1] != ']') 
		res->is_list = false;
	else {
		res->is_list = true;
		type[len-2] = '\0';
		len        -= 2;
	}
	
	for (i = 0; i < ARRAY_SIZE(DHCPD_TYPES); ++i) {
		if (strcmp(type, DHCPD_TYPES[i].name) != 0)
			continue;

		res->type = DHCPD_TYPES[i].type;
		return true;
	}
			
	fprintf(stderr, "unknown type '%s'\n", type);
	return false;
}

static bool dhcpd_parse_key(struct dhcpd_key_info *info, char *key)
{
	char				*pos = strrchr(key, '.');
	char				*endptr;
	long				code;
	struct dhcpd_key_str const	*key_info;
	struct dhcpd_key_str		key_info_tmp = { .code = 0 };

	if (!pos) 
		info->vendor = NULL;
	else {
		info->vendor = key;
		*pos         = '\0';
		key          = pos + 1;
	}

	if (!*key) {
		fprintf(stderr, "Empty <key>\n");
		return false;
	}

	code = strtol(key, &endptr, 10);

	if (*endptr != ':')
		key_info = dhcpd_parse_key_str(info->vendor, key);
	else if (*endptr == '\0')
		key_info = dhcpd_parse_key_int(info->vendor, code);
	else if (code <= 0 || code > 255) {
		fprintf(stderr, "Key #%lu out of range\n", code);
		return false;
	} else if (!dhcpd_parse_type(&key_info_tmp, endptr+1))
		return false;
	else {
		key_info_tmp.code = code;
		key_info          = &key_info_tmp;
	}

	if (!key_info)
		return false;

	info->code   = key_info->code;

	if (key_info->funcs)
		info->funcs = *key_info->funcs;
	else if (key_info->type != TYPE_SPECIAL)
		info->funcs = (key_info->is_list 
			       ? FUNCS_LIST[key_info->type] 
			       : FUNCS_NORM[key_info->type]);
	else {
		fprintf(stderr, "internal error: no function for special type defined\n");
		return false;
	}
	return true;
}

struct dhcpd_param * dhcpd_param_parse(char const *str_c)
{
	struct dhcpd_key_info	info;
	struct dhcpd_param	*res;
	char			*str = strdupa(str_c);
	char			*val;

	val = strchr(str, '=');
	if (!val) {
		fprintf(stderr, "No '=' in '<key>=<val>' pair: '%s'\n", str);
		return NULL;
	}

	*val++ = '\0';
	if (!dhcpd_parse_key(&info, str)) {
		fprintf(stderr, "failed to parse '%s'\n", str_c);
		return NULL;
	}

	res = malloc(sizeof(*res) + strlen(val));
	if (!res) {
		perror("malloc()");
		return NULL;
	}

	if (!info.funcs.fill_info(res, &info, val))
		goto err;

err:
	free(res);
	return NULL;
}
